#pragma once
#include "../util/gdix.h"	// windows.h, Rect, RANGE
#include "../util/vector.h"	// vector
#include "../util/smartpointer.h"
#include "../util/util.h"	// ASSERT

class Stream;

// Allgemeines Extent (terminal oder geschachtelt
class Extent{
public:
// Es wird davon ausgegangen dass:
// * die getXxx-Funktionen schnell ausgefhrt werden
// * der Wert von set() stets bei get() zurckkommt, auch dann, wenn's kleiner als getMin() und grer als getMax() ist
// Es erfolgt keine Absicherung gegen Multithreading und rekursive Aufrufe
 virtual bool set(int) =0;
 virtual bool setMin(int) =0;
 virtual int get() const =0;
 virtual int getMin() const =0;
 virtual int getMax() const =0;
 virtual int getWeight() const {return 0;}
 virtual void save(Stream&) {};
 virtual void load(Stream&) {};
};

#pragma warning(disable:4481)	// nonstandard override
// Terminale Extents
class ExtentFix:public Extent{
 int extent;	// Breite oder Hhe in Pixel, nichtnegativ
public:
 bool set(int x) override {if (extent==x) return false; ASSERT(!HIWORD(x)); extent=x; return true;}	// gewaltsam ndern
 bool setMin(int x) override {ASSERT(!HIWORD(x)); return set(x);}	// genauso
 int get() const override {return extent;}
 int getMin() const override {return extent;}	// auch wenn getMin() und getMax() etwas dagegen hatten
 int getMax() const override {return extent;}
 ExtentFix(int def=8):extent(def) {}
};

class ExtentVar:public ExtentFix{
 int minimum;
public:
 bool setMin(int x) override {if (minimum==x) return false; minimum=x; return true;}
 int getMin() const override {return minimum;}	// auch wenn getMin() und getMax() etwas dagegen hatten
 int getMax() const override {return 32766;}	// praktisch unbegrenzt
 int getWeight() const override {return 100;}	// Standardgewicht: Gleichberechtigung (absolut oder relativ??)
 ExtentVar(int min=16):ExtentFix(min),minimum(min) {}
};

// Nicht sonderlich performant, ohne Cache
class ExtentWndX:public Extent{
 HWND wnd;
 int weight;
public:
 bool set(int x) override {Rect<>r(wnd); if (r.width()==x) return false; r.right=x; r.bottom-=r.top; SetWindowPos(wnd,0,0,0,r.right,r.bottom,SWP_NOMOVE|SWP_NOZORDER); return true;}
 int get() const override {Rect<>r(wnd); return r.width();}
 int getMin() const override {MINMAXINFO mmi; SendMessage(wnd,WM_GETMINMAXINFO,0,(LPARAM)&mmi); return mmi.ptMinTrackSize.x;}
 int getMax() const override {MINMAXINFO mmi; SendMessage(wnd,WM_GETMINMAXINFO,0,(LPARAM)&mmi); return mmi.ptMaxTrackSize.x;}
 int getWeight() const override {return weight;}
 ExtentWndX(HWND w,int we=100):wnd(w),weight(we) {}
};


// Extents in einem Vektor (also hintereinanderweg hnlich struct)
class ExtentList:public Extent{
 friend class GridLayout;
 friend class PlotXY;
 typedef vector<uniq_ptr<Extent> >Extents;
 Extents extents;	// Fr PlotXY stets ungerade und >=3 Elemente: Linker/oberer Rand, Plotbereich, rechter/unterer Rand
 class Cache {
 public:
  struct Entry{
   double x,min,max,weight;
   double delta_to_keep_limit(double newsize) const;
  };
  typedef vector<Entry> Entries;
  Entries entries;
  bool load(const Extents&src,double&wg);	// false wenn's hinreichend gleich war
  void commit(Extents&dst);
 }cache;
 char makeValidIndex(char i) const {if ((size_t)i>=extents.size()) i=(char)extents.size()-1; return i;}
 char makeValidSpan(char valid_i, char span) const;
 Extents::const_iterator locatePos(int&) const;
 RANGE<int>extent(char a,char span)const;
 typedef int (Extent::*cb_t)() const;	// Eine der vier getXxx-Funktionen in Extent. Tatschlich geht der Aufruf ber die VMT des aktuellen Objekts!
 int getSum(cb_t) const;
 int get() const override {return getSum(&Extent::get);}
 bool set(int) override;		// Legt Verteilung (neu) fest, bei WM_SIZE
 void setMin(const int*);	// Legt Minima fest
 bool setMin(int) override {return false;}
 int getMin() const override {return getSum(&Extent::getMin);}	// fr WM_GETMINMAXINFO
 int getMax() const override {return getSum(&Extent::getMax);}	// fr WM_GETMINMAXINFO
 int getWeight() const override {return getSum(&Extent::getWeight);}
 char draggable(int) const;	// liefert Spaltenindex oder -1 wenn nicht draggable
 void drag(char i, int delta);	// Verschiebt das Layout
 void setsize(char i,int sz);	// Setzt Breite/Hhe eines Einzelelements
// ExtentList(int def=100);	// Generiert das Minimum: 1 Element
 bool insert(char index,Extent*);	// Zeilen/Spalte hinzufgen, index=-1: Am Ende
 bool erase(char index);	// index=-1: Letztes Element, -2 = vorletztes usw.
 void debuglog() const;
};
#pragma warning(default:4481)

// Zwei orthogonale Extent-Vektoren
class GridLayout{
public:
 typedef Point<char> gridpos_t;
 struct gridrange_t{
  char x,y,colspan,rowspan;
 };
 ExtentList extents_xy[2];
 Rect<>extent(gridrange_t) const;
 Rect<>extent(char ix, char iy, char colspan=1, char rowspan=1) const{
  gridrange_t gr={ix,iy,colspan,rowspan};
  return extent(gr);
 }
 Rect<>fullrect() const {return extent(0,0,-1,-1);}
 char gridsize(unsigned xysel) const {return static_cast<char>(extents_xy[xysel&1].extents.size());}
};
